# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2013 Vifib SARL and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

import slapos.slap

import logging
import os
import subprocess
import sys
import time
import urllib2

UNIT_TEST_ERP5TESTNODE = 'UnitTest'

class ResiliencyTestSuite(object):
  """
  Abstract class supposed to be extended by Resiliency Test Suites.
  """
  def __init__(self,
               server_url, key_file, cert_file,
               computer_id, partition_id, software,
               namebase,
               root_instance_name,
               sleep_time_between_test=900,
               total_instance_count="2",
               type=None):
    self.server_url = server_url
    self.key_file = key_file
    self.cert_file = cert_file
    self.computer_id = computer_id
    self.partition_id = partition_id
    self.software = software
    self.namebase = namebase
    self.total_instance_count = total_instance_count
    self.root_instance_name = root_instance_name
    self.sleep_time_between_test = sleep_time_between_test
    self.test_type = type

    slap = slapos.slap.slap()
    slap.initializeConnection(server_url, key_file, cert_file)
    self.partition = slap.registerComputerPartition(
        computer_guid=computer_id,
        partition_id=partition_id
    )

    self.logger = logging.getLogger('SlaprunnerResiliencyTest')
    self.logger.setLevel(logging.DEBUG)

  def _doTakeover(self, namebase, target_clone):
    """
    Private method.
    Make the specified clone instance takeover the main instance.
    """
    self.logger.info('Replacing main instance by clone instance %s%s...' % (
        self.namebase, target_clone))

    # Request
    root_partition_parameter_dict = self._getPartitionParameterDict()
    takeover_url = root_partition_parameter_dict['takeover-%s-%s-url' % (namebase, target_clone)]
    takeover_password = root_partition_parameter_dict['takeover-%s-%s-password' % (namebase, target_clone)]
    # Connect to takeover web interface
    takeover_page_content = urllib2.urlopen(takeover_url).read()
    # Wait for importer script to be not running
    while 'Importer script(s) of backup in progress: True' in takeover_page_content:
      time.sleep(10)
      takeover_page_content = urllib2.urlopen(takeover_url).read()
    # Do takeover
    takeover_result = urllib2.urlopen('%s?password=%s' % (takeover_url, takeover_password)).read()
    if 'Error' in takeover_result:
      raise Exception('Error while doing takeover: %s' % takeover_result)

    self.logger.info('Done.')

  def generateData(self):
    """
    Generate data that will be used by the test.
    """
    raise NotImplementedError('Overload me, I am an abstract method.')

  def pushDataOnMainInstance(self):
    """
    Push our data to the main instance.
    """
    raise NotImplementedError('Overload me, I am an abstract method.')

  def checkDataOnCloneInstance(self):
    """
    Check that, on the ex-clone, now-main instance, data is the same as
    what we pushed to the ex-main instance.
    """
    raise NotImplementedError('Overload me, I am an abstract method.')

  def deleteTimestamp():
    """
    XXX-Nicolas delete .timestamp in test partition to force the full processing
    by slapgrid, to force the good parameters to be passed to the instances of the tree
    """
    home = os.getenv('HOME')
    timestamp = os.path.join(home, '.timestamp')
    os.remove(timestamp)

  def _getPartitionParameterDict(self):
    """
    Helper.
    Return the partition parameter dict of the main root ("resilient") instance.
    """
    return self.partition.request(
        software_release=self.software,
        software_type='resilient',
        partition_reference=self.root_instance_name
    ).getConnectionParameterDict()
    self.deleteTimestamp()

  def _returnNewInstanceParameter(self, parameter_key, old_parameter_value):
    """
    Helper, can be used inside of checkDataOnCloneInstance.
    Wait for the new parameter (of old-clone new-main instance) to appear.
    Check than it is different from the old parameter
    """
    # if we are inside of a classical erp5testnode: just return the same parameter.
    if self.test_type == UNIT_TEST_ERP5TESTNODE:
      return old_parameter_value

    self.logger.info('Waiting for new main instance to be ready...')
    new_parameter_value = None
    while not new_parameter_value or new_parameter_value == 'None' or  new_parameter_value == old_parameter_value:
      self.logger.info('Not ready yet. SlapOS says new parameter value is %s' % new_parameter_value)
      new_parameter_value = self._getPartitionParameterDict().get(parameter_key, None)
      time.sleep(30)
    self.logger.info('New parameter value of instance is %s' % new_parameter_value)
    return new_parameter_value

  def _testClone(self, clone):
    """
    Private method.
    Launch takeover and check for a specific clone.
    """
    self._doTakeover(self.namebase, clone)
    self.logger.info('Testing %s%s instance.' % (self.namebase, clone))

    # Wait for XX minutes so that replication is done
    self.logger.info(
      'Sleeping for %s seconds before testing clone %s.' % (
          self.sleep_time_between_test,
          clone
        ))
    time.sleep(self.sleep_time_between_test)

    if self.test_type == UNIT_TEST_ERP5TESTNODE: # Run by classical erp5testnode using slapproxy
      # Run manually slapos node instance
      # XXX hardcoded path
      self.logger.info('Running "slapos node instance"...')
      slapos_configuration_file_path = os.path.join(
          os.path.dirname(sys.argv[0]),
          '..', '..', '..', 'slapos.cfg'
      )
      # Output is huge and we don't want to store it in memory nor print it
      devnull = open('/dev/null', 'w')
      command = ['/opt/slapos/bin/slapos', 'node', 'instance',
                 '--cfg=%s' % slapos_configuration_file_path,
                 '--pidfile=slapos.pid']
      subprocess.Popen(command, stdout=devnull, stderr=devnull).wait()
      subprocess.Popen(command, stdout=devnull, stderr=devnull).wait()
      subprocess.Popen(command, stdout=devnull, stderr=devnull).wait()

    success = self.checkDataOnCloneInstance()
    if success:
      return True
    return False

  def runTestSuite(self):
    """
    Generate data to send,
    Push data on main instance,
    Wait for replication to be done,
    For each clone: Do a takeover, Check data.
    """
    self.generateData()

    self.pushDataOnMainInstance()

    # In resilient stack, main instance (example with KVM) is named "kvm0",
    # clones are named "kvm1", "kvm2", ...
    clone_count = int(self.total_instance_count) - 1
    # So first clone starts from 1.
    current_clone = 1

    # In case we have only one clone: test the takeover twice
    # so that we test the reconstruction of a new clone.
    if clone_count == 1:
      return self._testClone(1)
      #for i in range(2):
      #  success = self._testClone(1)
      #  if not success:
      #    return False

    else:
      # Test each clone
      while current_clone <= clone_count:
        success = self._test_clone(current_clone)
        if not success:
          return False
        current_clone = current_clone + 1

    # All clones have been successfully tested: success.
    return True




import argparse
import csv
import feedparser
import httplib # To avoid magic numbers
import io
import json
import logging
import math
import os
import socket
import sys
import time
from datetime import datetime
from hashlib import sha512

from atomize import Entry
from atomize import Feed
from atomize import Content
from flask import Flask
from flask import abort
from flask import request
app = Flask(__name__)

# csv entries can be very large, increase limit.
csv.field_size_limit(sys.maxsize)

@app.route('/get/<feed>')
def get_feed(feed):
  global app
  feedpath = os.path.join(app.config['FEEDS'], feed)
  if not os.path.exists(feedpath):
    abort(httplib.NOT_FOUND)

  # XXX: Add a way to specify a title
  feed_title = 'Untitled'
  feed_guid = request.url
  # XXX: Add a way to specify an author
  feed_author = 'No author'

  entries = []
  feed_updated = 0
  with open(feedpath, 'r') as feed_file:
    reader = csv.reader(feed_file)
    for row in reader:
      timestamp, title, content, guid = row

      timestamp = int(timestamp)
      # Keep the maximum timestamp without
      # looping once more.
      if timestamp > feed_updated:
        feed_updated = timestamp

      entries.append(Entry(title=title,
                           guid=guid,
                           updated=datetime.fromtimestamp(timestamp),
                           content=Content(content, content_type='html'),
                          ))
  entries.reverse()
  feed = Feed(title=feed_title,
              updated=datetime.fromtimestamp(feed_updated),
              guid=feed_guid,
              author=feed_author,
              entries=entries,
              self_link=request.url)

  return (feed.feed_string(),
          httplib.OK,
          {'Content-Type': 'application/atom+xml'}
         )

@app.route('/notify/<int:transaction_id>', methods=['POST'])
def notify(transaction_id):
  global app
  try:
    feed = feedparser.parse(request.data)
  except ValueError:
    abort(httplib.BAD_REQUEST)

  if feed.bozo: # Malformed XML
    abort(httplib.BAD_REQUEST)

  try:
    callback_filepath = os.path.join(app.config['CALLBACKS'],
                                     sha512(str(feed.feed.id)).hexdigest())
    if not os.path.exists(callback_filepath):
      abort(httplib.NOT_FOUND)
  except AttributeError:
    abort(httplib.BAD_REQUEST)


  abort_it = False

  for callback in io.open(callback_filepath, 'r', encoding='utf8'):
    timestamp = int(math.floor(time.mktime(feed.feed.updated_parsed)))

    equeue_request = json.dumps({
        'command': '%s\0--transaction-id\0%s' % (callback, transaction_id),
        'timestamp': timestamp,
        })

    equeue_socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    equeue_socket.connect(app.config['EQUEUE_SOCKET'])
    equeue_socket.send(equeue_request)
    result = equeue_socket.recv(len(callback))
    equeue_socket.close()

    if result != callback:
      abort_it = True

  if abort_it:
    # XXX if possible, communicate info about the failed callbacks
    abort(httplib.INTERNAL_SERVER_ERROR)

  return '', httplib.NO_CONTENT


def getLogger(logfile):
  logger = logging.getLogger("Notifier")
  logger.setLevel(logging.INFO)
  handler = logging.FileHandler(logfile)
  # Natively support logrotate
  formatter = logging.Formatter('%(asctime)s %(levelname)s: %(message)s')
  handler.setFormatter(formatter)
  logger.addHandler(handler)
  return logger

def main():
  global app
  parser = argparse.ArgumentParser(description="Atom server")
  parser.add_argument('-c', '--callbacks', nargs=1, required=True,
                      help="Callback directory.")
  parser.add_argument('-f', '--feeds', nargs=1, required=True,
                      help="Feeds directory")
  parser.add_argument('-s', '--equeue-socket', dest='equeue_socket',
                      nargs=1, required=True, help="EQUEUE Server socket")
  parser.add_argument('-l', '--logfile', nargs=1, required=False,
                      help="Logfile.", default='', type=str)
  parser.add_argument('host', metavar='hostname', default='0.0.0.0', nargs='?')
  parser.add_argument('port', metavar='port', type=int, default=8080, nargs='?')
  args = parser.parse_args()
  app.config.update(FEEDS=args.feeds[0],
                    CALLBACKS=args.callbacks[0],
                    EQUEUE_SOCKET=args.equeue_socket[0])
  # Set log
  logfile = args.logfile[0]
  if logfile:
    logger = getLogger(logfile)
    app.logger.addHandler(logger)
    app.logger.info('starting server')
  app.run(host=args.host, port=args.port)

if __name__ == '__main__':
  main()



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

# dependency: python2-pycurl

import sys
import tempfile
import os
import argparse
import ConfigParser
import re

import pycurl

from mimetools import Message
from cStringIO import StringIO
from HTMLParser import HTMLParser

begins_by_known_protocol_re = re.compile("^https?://")
get_protocol_re = re.compile("^([a-z]+)://")

def checkWebpageHttpCacheHit(url_list, resolve_list=[], cookie_jar_path=None):
  report_line_list = []
  class MyHTMLParser(HTMLParser):
    def __init__(self, base_url):
      HTMLParser.__init__(self)
      self.base_url = base_url
    def handle_starttag(self, tag, attrs):
      def handle_url(url):
        if url.startswith("//"):
          protocol = get_protocol_re.match(self.base_url)
          if protocol is not None:
            protocol = protocol.group(1)
          else:
            protocol = "http"
          url = protocol + ":" + url
        elif begins_by_known_protocol_re.match(url) is None:
          url = self.base_url + ("" if self.base_url.endswith("/") else "/") + url
        do_request(url)
      # handle images and js scripts
      if tag == 'img' or tag == 'script':
        for attr in attrs:
          if attr[0] == 'src':
            handle_url(attr[1])
      # handle css and favicon
      if tag == 'link' and \
         any(attr[0] == "rel" and attr[1] in ("stylesheet", "shortcut icon") for attr in attrs):
        for attr in attrs:
          if attr[0] == "href":
            handle_url(attr[1])

  headers = {
    "User-Agent": "test",
    "Accept-Encoding": "gzip, deflate",
    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
    "Accept-Language": "en-US,en;q=0.5",
    # "Connection": "keep-alive"
  }

  parsed_url_dict = set()
  def do_request(url):
    if url in parsed_url_dict: return
    parsed_url_dict.add(url)
    print("Checking cache hit for " + url)
    c = pycurl.Curl()
    response_headers = StringIO()
    output = StringIO()
    c.setopt(c.URL, url)
    c.setopt(c.RESOLVE, resolve_list)
    c.setopt(c.WRITEFUNCTION, output.write)
    c.setopt(c.HEADERFUNCTION, response_headers.write)
    c.setopt(c.COOKIEFILE, cookie_jar_path)
    c.setopt(c.COOKIEJAR, cookie_jar_path)
    try:
      c.perform() # perform a request before testing if the cache is hit
      response_headers.truncate(0)
      output.truncate(0)
      c.perform()
      if str(c.getinfo(pycurl.HTTP_CODE))[0:1] != "2":
        if c.getinfo(pycurl.HTTP_CODE) >= 400:
          report_line_list.append("Status code %s received for %s" % (c.getinfo(pycurl.HTTP_CODE), url))
        else:
          print("Status code %s not handled" % c.getinfo(pycurl.HTTP_CODE))
        return
    except pycurl.error as e:
      if e[0] == 6: # Could not resolve domain
        report_line_list.append(e[1])
        return
      raise

    response_headers.seek(0)
    status = response_headers.readline().split(" ")[1]

    m = Message(response_headers)

    # see http://labs.omniti.com/people/mark/ats_sa/slides.html#slide-18
    if not any("[cHs" in header for header in m.getheaders('via')) and \
       not any("HIT" in header for header in m.getheaders("x-cache")):
      report_line_list.append("No cache hit found for " + url)
      # report_line_list.append(response_headers.getvalue().rstrip())

    if m.getheader("content-type", "").startswith("text/html"):
      MyHTMLParser(url).feed(output.getvalue())

    response_headers.close()
    output.close()

  for url in url_list:
    do_request(url)

  if report_line_list == []:
    print("No failure.")
    return True
  print("Failure report:")
  for line in report_line_list:
    print("- " + line)
  return False

def getConfig(config_parser, section, option, default=None, raw=False, vars=None):
  try:
    return config_parser.get(section, option, raw=raw, vars=vars)
  except ConfigParser.NoOptionError:
    return default

def main():
  parser = argparse.ArgumentParser()
  parser.add_argument("-K", "--config", nargs=1, metavar="FILE", help="read config from FILE")
  parser.add_argument("-c", "--cookie-jar", nargs=1, metavar="FILE", help="write cookies to FILE after operation")
  parser.add_argument("--resolve", nargs="+", default=[], metavar="HOST:PORT:ADDRESS", help="force resolve of HOST:PORT to ADDRESS")
  parser.add_argument("url-list", nargs="*", metavar="URL", default=[])
  args = parser.parse_args()
  args.url_list = getattr(args, "url-list")

  if args.config is not None:
    parser = ConfigParser.ConfigParser()
    parser.read(args.config)
    if args.url_list == []:
      args.url_list = getConfig(parser, "public", "url-list", "").split()
    if args.resolve == []:
      args.resolve = getConfig(parser, "public", "resolve-list", "").split()
    if args.cookie_jar is not None:
      args.cookie_jar = getConfig(parser, "public", "cookie-jar-filename", "")
      if args.cookie_jar == "":
        args.cookie_jar = None

  if args.url_list == []:
    raise ValueError("No url to check")

  if args.cookie_jar is None:
    cookie_jar = tempfile.NamedTemporaryFile(delete=False)
    cookie_jar_path = cookie_jar.name
    cookie_jar.close()
  else:
    cookie_jar_path = args.cookie_jar
  try:
    res = checkWebpageHttpCacheHit(args.url_list, args.resolve, cookie_jar_path)
  finally:
    if args.cookie_jar is None:
      os.unlink(cookie_jar_path)
  return 0 if res else 1

if __name__ == "__main__":
  sys.exit(main())



import ZODB.FileStorage
import ZODB.serialize
import argparse
import time
import traceback

def run():
  now = time.time()
  parser = argparse.ArgumentParser(description='ZODB offline packer')
  parser.add_argument('files', metavar='F', type=str, help='Files to pack',
    nargs='+')
  parser.add_argument('--days', '-d', type=float,
    help='Amount of days, from now, to pack')

  args = parser.parse_args()

  point = now - (3600 * 24 * args.days)

  print 'Now is %s' % time.asctime(time.localtime(now))
  print 'Will pack until %s' % time.asctime(time.localtime(point))
  failures = 0
  for f in args.files:
    b = time.time()
    print 'Trying to pack %r' % f
    try:
      pack(point, f)
    except Exception:
      print 'Failed to pack %r:' % f
      traceback.print_exc()
      failures += 1
    print 'Finished %s in %.3fs' % (f, time.time() - b)
  if failures:
    print 'Failed files: %s' % failures
    return failures
  else:
    print 'All files sucessfully packed.'
    return 0

def pack(point, f):
  ZODB.FileStorage.FileStorage(f).pack(point, ZODB.serialize.referencesf)



import socket
import logging
import time
import ConfigParser
import logging.handlers
import urllib2
import subprocess
import re
import sys
import shutil
import netifaces
import random

botname = socket.gethostname()

# rtt min/avg/max/mdev = 1.102/1.493/2.203/0.438 ms
ping_re = re.compile(
    ".*"
    "(?P<min>[\d\.]+)/"
    "(?P<max>[\d\.]+)/"
    "(?P<avg>[\d\.]+)/"
    "(?P<mdev>[\d\.]+) ms"
    )

date_reg_exp = re.compile('\d{4}[-/]\d{2}[-/]\d{2}')

def _get_network_gateway(self):
  return netifaces.gateways()["default"][netifaces.AF_INET][0]

def _test_dns(name):
  begin = time.time()
  try:
    socket.gethostbyname(name)
    resolution = 200
    status = "OK"
  except socket.gaierror:
    resolution = 600
    status = "Cannot resolve the hostname"

  resolving_time = time.time() - begin
  return ('DNS', name, resolution, resolving_time, status)

def _test_ping(host, timeout=10):
  proc = subprocess.Popen(('ping', '-c', '4', '-w', str(timeout), host), 
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE)

  out, err = proc.communicate()
  packet_loss_line, summary_line = (out.splitlines() or [''])[-2:]
  m = ping_re.match(summary_line)
  info_list = ('PING', host, '600', 'failed', "Cannot ping host")
  if ' 0% packet loss' in packet_loss_line:
    if m:
      info_list = ('PING', host, '200', m.group('avg'),
           'min %(min)s max %(max)s avg %(avg)s' % m.groupdict())

  return info_list

def _test_url_request(url, request):
  begin = time.time()
  try:
    response = urllib2.urlopen(request)
    if url != response.url:
      raise RuntimeError, 'Redirected to %s' % response.url
    response.read()
    rendering_time = time.time() - begin
    info_list = ('GET', url, response.code, rendering_time, "OK")
  except urllib2.HTTPError, e: # 40x errors
    e.fp.read()
    rendering_time = time.time() - begin
    info_list = ('GET', url, e.code, rendering_time, "HTTPError")
  except urllib2.URLError, e: # DNS failure
    rendering_time = time.time() - begin
    info_list = ('GET', url, '600', rendering_time, "DNS Failure")
  except RuntimeError, e:
    info_list = ('GET', url, '600', 'failed')

  return info_list

def is_rotate_log(log_file_path):
  try:
    log_file = open(log_file_path, 'r')
  except IOError:
    return False

  # Handling try-except-finally together.
  try:
    try:
      log_file.seek(0, 2)
      size = log_file.tell()
      log_file.seek(-min(size, 4096), 1)
      today = time.strftime("%Y-%m-%d")
      
      for line in reversed(log_file.read().split('\n')):
        if len(line.strip()):
          match_list = date_reg_exp.findall(line)
          if len(match_list):
            if match_list[0] != today:
              return ValueError(match_list[0])
            break

    except IOError:
      return False
  finally:
    log_file.close()

def create_logger(name, log_folder):
  new_logger = logging.getLogger(name)

  new_logger.setLevel(logging.DEBUG)
  log_file = '%s/network_bench.%s.log' % (log_folder, name)
  handler = logging.handlers.TimedRotatingFileHandler(
                     log_file, when="D",
                     backupCount=1000)

  last_date = is_rotate_log(log_file)
  if last_date:
    handler.doRollover()
    today = time.strftime("%Y-%m-%d")
    shutil.move("%s.%s" % (log_file, today),
                "%s.%s" % (log_file, last_date))
    # The simpler the better
    sp = subprocess.Popen("gzip %s.%s" % (log_file, last_date),
               stdout=subprocess.PIPE, stdin=subprocess.PIPE, shell=True)
    sp.communicate()

  format = "%%(asctime)-16s;%%(levelname)s;%s;%%(message)s" % botname
  handler.setFormatter(logging.Formatter(format))
  new_logger.addHandler(handler)
  return new_logger

def main():
  if len(sys.argv) not in [2, 3]:
    print " USAGE: %s configuration_file [log_folder]" % sys.argv[0]
    return

  config = ConfigParser.ConfigParser()
  config.read(sys.argv[1])

  if len(sys.argv) == 3:
    log_folder = sys.argv[2]
  else:
    log_folder = "."

  delay = random.randint(0, 30)

  name_list = config.get("network_bench", "dns")
  url_list = config.get("network_bench", "url")
  ping_list = config.get("network_bench", "ping")

  logger = create_logger("info", log_folder)

  logger.debug('Starting a new test in %s seconds' % delay)

  time.sleep(delay)

  for name in name_list.split():
    info_list = _test_dns(name)
    logger.info(';'.join(str(x) for x in info_list))

  # ping
  for host in ping_list.split():
    info_list = _test_ping(host)
    logger.info(';'.join(str(x) for x in info_list))

  # http
  for url in url_list.split():
    info_list = _test_url_request(url, urllib2.Request(url))
    logger.info(';'.join(str(x) for x in info_list))

